extern crate rand;

use std::time::Instant;
use rand::{Rng, SeedableRng, StdRng};

const DIMENSION: usize = 2;

#[derive(Debug, Copy, Clone)]
struct Vertex {
    xi : [f64; DIMENSION]
}

#[allow(dead_code)]
fn neighbor_step(rng: &mut StdRng, v: Vertex) -> Vertex {
    let step: f64 = 0.01;

    let mut n = v;

    let i = rng.gen::<usize>() % n.xi.len();
    if rng.gen::<bool>() {
        n.xi[i] += step;
    } else {
        n.xi[i] -= step;
    }
    return n;
}

#[allow(dead_code)]
fn griewank(v: Vertex) -> f64 {
    let mut add = 1f64;
    let mut mult = 1f64;
    for (i, x) in v.xi.iter().enumerate() {
        add += (1f64/4000f64)*(x.powi(2));
        mult *= ((1f64/((i + 1) as f64)) * x * ((i + 1) as f64).sqrt()).cos();
    }
    return add - mult;
}

#[allow(dead_code)]
fn circle(v: Vertex) -> f64 {
    let mut y = 0f64;
    for x in v.xi.iter() {
        y += x.powi(2);
    }
    return y;
}

fn ord_f64(a: f64, b: f64) -> bool {
    return a < b;
}

fn lahc<T: std::fmt::Debug + std::marker::Copy, U: std::fmt::Display + std::marker::Copy>(
    init: T,
    memory_size: usize,
    search_for: u64,
    func: &Fn(T) -> U,
    ord: &Fn(U, U) -> bool,
    neighbor: &Fn(&mut StdRng, T) -> T,
    rng: &mut StdRng,    
    debug: bool
) -> T {
    let mut min: T = init;
    let mut memory = vec![init; memory_size];
    let mut n = 0u32;

    let start = Instant::now();
    while start.elapsed().as_secs() < search_for {
        let x = memory[(n % (memory_size as u32)) as usize];
        let xv = neighbor(rng, x);

        if ord(func(xv), func(x)) {
            memory[(n % (memory_size as u32)) as usize] = xv;
        }

        if ord(func(memory[(n % (memory_size as u32)) as usize]), func(min)) {
            min = memory[(n % (memory_size as u32)) as usize];
            if debug {
                println!("f({:?}) = {}", min, func(min));
            }
        }
        n += 1;
    }
    return min;
}

fn main() {
    const DEBUG: bool = false;

    let mut rng: StdRng;
    if DEBUG {
        let seed: &[_] = &[1,0,1,2,5];
        rng = SeedableRng::from_seed(seed);
    } else {
        if let Ok(new_rng) = StdRng::new(){
            rng = new_rng;
        } else {
            panic!("IO error");
        }
    }

    let func = &circle;
    let init: Vertex = Vertex {xi : [5f64; DIMENSION]};
    let min = lahc::<Vertex, f64>(init, 100, 5, func, &ord_f64, &neighbor_step,&mut rng, DEBUG);
    println!("f({:?}) = {}", min, func(min));
}
