extern crate rand;

use rand::{Rng, SeedableRng, StdRng};

const DIMENSION: usize = 10;

#[derive(Debug, Copy, Clone)]
struct Vertex {
    xi : [f64; DIMENSION]
}

#[derive(Debug, Copy, Clone)]
struct Knapsack {
    xi : [bool; DIMENSION]
}


fn metropolis(rng: &mut StdRng, delta_f : f64, temperature: f64) -> bool {
    return (delta_f <= 0f64) || rng.gen::<f64>() < (-delta_f/temperature).exp();
}

fn simulated_annealing<T: std::fmt::Debug + std::marker::Copy, U: std::cmp::PartialOrd + std::fmt::Display + std::marker::Copy>(
    start: T,
    temperature_range: (f64, f64),
    rng: &mut StdRng,
    cooling: &Fn(f64) -> f64,
    func: &Fn(T) -> U,
    delta: &Fn(U, U) -> f64,
    neighbor: &Fn(&mut StdRng, T) -> T,
    debug: bool
) -> T {
    let mut x: T = start;
    let mut min: T = x;
    let mut fmin: U = func(min);

    let mut temperature = temperature_range.0;
    let min_temperature: f64 = temperature_range.1;

    while temperature > min_temperature {
        if debug {
            println!("T° = {}", temperature);
        }
        let mut it_since_last_min : u32 = 0;
        while it_since_last_min < 500_000 {
            let xv = neighbor(rng, x);

            let fx = func(x);
            let fxv = func(xv);

            let delta_f : f64 = delta(fxv, fx);
            if metropolis(rng, delta_f, temperature) {
                if delta_f <= 0f64 && fxv <= fmin {
                    min = xv;
                    fmin = fxv;
                    if delta_f < 0f64 && fxv < fmin {
                        it_since_last_min = 0;
                    }

                    if debug {
                        println!("f({:?})\n= {}; delta: {}", min, fmin, delta_f);
                    }
                }
                x = xv;
            }
            it_since_last_min += 1;
        }
        temperature = cooling(temperature);
    }

    return min;
}

// Continuous function

#[allow(dead_code)]
fn neighbor_step(rng: &mut StdRng, v: Vertex) -> Vertex {
    let step: f64 = 0.0995;

    let mut n = v;

    let i = rng.gen::<usize>() % n.xi.len();
    if rng.gen::<bool>() {
        n.xi[i] += step;
    } else {
        n.xi[i] -= step;
    }
    return n;
}

#[allow(dead_code)]
fn griewank(v: Vertex) -> f64 {
    let mut add = 1f64;
    let mut mult = 1f64;
    for (i, x) in v.xi.iter().enumerate() {
        add += (1f64/4000f64)*(x.powi(2));
        mult *= ((1f64/((i + 1) as f64)) * x * ((i + 1) as f64).sqrt()).cos();
    }
    return add - mult;
}

#[allow(dead_code)]
fn circle(v: Vertex) -> f64 {
    let mut y = 0f64;
    for x in v.xi.iter() {
        y += x.powi(2);
    }
    return y;
}

// Discret

#[allow(dead_code)]
fn neighbor_knapsack(rng: &mut StdRng, v: Knapsack) -> Knapsack {
    let mut n = v;
    let i = rng.gen::<usize>() % n.xi.len();
    n.xi[i] = !n.xi[i];
    
    return n;
}

#[allow(dead_code)]
fn knapsack_value(k: Knapsack) -> f64 {
    const PRICES : [f64; DIMENSION] = [92f64,57f64,49f64,68f64,60f64,43f64,67f64,84f64,87f64,72f64];
    const WEIGHT : [u16; DIMENSION] = [23,31,29,44,53,38,63,85,89,82];
    const MAX_WEIGHT : u16 = 165;

    let mut v: f64 = 0f64;
    let mut w: u16 = 0;
    for (i, x) in k.xi.iter().enumerate() {
        if *x {
            v += PRICES[i];
            w += WEIGHT[i];
        }
    }

    if w > MAX_WEIGHT {
        v = 0f64;
    }
    return -v;
}


// other
fn cooling(t: f64) -> f64 {
    return t / 1.05f64;
}

fn delta(a: f64, b: f64) -> f64 {
    return a - b;
}

fn main() {
    const DEBUG: bool = false;

    let mut rng: StdRng;
    if DEBUG {
        let seed: &[_] = &[1,0,1,2,5];
        rng = SeedableRng::from_seed(seed);
    } else {
        if let Ok(new_rng) = StdRng::new() {
            rng = new_rng;
        } else {
            panic!("IO error");
        }
    }

    let func = &knapsack_value;
    let start: Knapsack = Knapsack {xi : [false; DIMENSION]};

    let min = simulated_annealing::<Knapsack, f64>(start, (100_000f64, 0.05f64), &mut rng, &cooling, func, &delta, &neighbor_knapsack, DEBUG);
    println!("min : {:?}", min);
    println!("f(min): {}", func(min));

}
