extern crate rand;

use std::time::Instant;
use std::ops::{Add, Sub, Mul};
use rand::{Rng, SeedableRng, StdRng};
use rand::distributions::{Range, Sample};

const DIMENSION: usize = 5;

#[derive(Debug, Copy, Clone)]
struct Vector {
    xi: [f64; DIMENSION],
}

impl Add for Vector {
    type Output = Vector;

    fn add(self, other: Vector) -> Vector {
        let mut result: Vector = self;
        for i in 0..DIMENSION {
            result.xi[i] += other.xi[i];
        }
        return result;
    }
}

impl Sub for Vector {
    type Output = Vector;

    fn sub(self, other: Vector) -> Vector {
        let mut result: Vector = self;
        for i in 0..DIMENSION {
            result.xi[i] -= other.xi[i];
        }
        return result;
    }
}

impl Mul<f64> for Vector {
    type Output = Vector;

    fn mul(self, rhs: f64) -> Vector {
        let mut result: Vector = self;
        for i in 0..DIMENSION {
            result.xi[i] *= rhs;
        }
        return result;
    }
}

#[derive(Debug, Copy, Clone)]
struct Particule {
    pos: Vector,
    speed: Vector,
    min: Vector,
}

#[allow(dead_code)]
fn circle(v: Vector) -> f64 {
    let mut y = 0f64;
    for x in v.xi.iter() {
        y += x.powi(2);
    }
    return y;
}

#[allow(dead_code)]
fn griewank(v: Vector) -> f64 {
    let mut add = 1f64;
    let mut mult = 1f64;
    for (i, x) in v.xi.iter().enumerate() {
        add += (1f64/4000f64)*(x.powi(2));
        mult *= ((1f64/((i + 1) as f64)) * x * ((i + 1) as f64).sqrt()).cos();
    }
    return add - mult;
}


fn swarm(
    func: &Fn(Vector) -> f64,
    rng: &mut StdRng,
    search_for: u64,
    debug: bool,  
) -> Vector {
    let size: usize = 10 + (2f64 * (DIMENSION as f64).sqrt()).round() as usize;
    if debug {
        println!("size = {}", size);
    }
    let mut particules = Vec::with_capacity(size);
    for _i in 0..size {
        let mut init: Vector = Vector { xi: [0f64; DIMENSION] };
        let mut speed: Vector = Vector { xi: [0f64; DIMENSION] };

        let mut value = Range::new(-1e10, 1e10);
        let mut unit = Range::new(-1f64, 1.);

        let mut speed_norm = 0f64;
        for i in 0..DIMENSION {
            init.xi[i] = value.sample(rng);
            speed.xi[i] = unit.sample(rng);

            speed_norm += (speed.xi[i]).powi(2);
        }
        speed_norm = speed_norm.sqrt();

        for x in speed.xi.iter_mut() {
            *x /= speed_norm
        }

        particules.push(Particule {
            pos: init,
            speed: speed,
            min: init,
        });
    }

    const PHI: f64 = 0.8f64;
    const C: f64 = 1.62f64;
    
    let mut min: Vector = particules[0].pos;

    let start = Instant::now();
    while start.elapsed().as_secs() < search_for {
        let mut current_min = min;
        for particule in particules.iter_mut() {
            particule.speed = particule.speed * PHI + (particule.min - particule.pos) * (C * rng.gen::<f64>()) + (min - particule.pos) * (C * rng.gen::<f64>());
            particule.pos = particule.pos + particule.speed;

            if func(particule.pos) < func(particule.min) {
                particule.min = particule.pos;

                if func(particule.min) < func(current_min) {
                    current_min = particule.min;
                    if debug {
                        println!("f({:?}) = {}", min, func(min));
                    }
                }
            }

        }
        min = current_min;
    }

    return min;
 }

fn main() {
    const DEBUG: bool = false;

    let mut rng: StdRng;
    if DEBUG {
        let seed: &[_] = &[1,0,1,2,5];
        rng = SeedableRng::from_seed(seed);
    } else {
        if let Ok(new_rng) = StdRng::new() {
            rng = new_rng;
        } else {
            panic!("IO error");
        }
    }

    let func = griewank;

    const N: usize = 10;
    const SEARCH_FOR: u64 = 1;

    let mut min: Vector = swarm(&func, &mut rng, SEARCH_FOR, DEBUG);
    for _i in 0..(N-1) {
        let local_min = swarm(&func, &mut rng, SEARCH_FOR, DEBUG);
        if func(local_min) < func(min) {
            min = local_min
        }
    }
    println!("f({:?}) = {}", min, func(min));
}
