use std::collections::BinaryHeap;
use std::collections::HashSet;
use std::cmp::Ordering;
use std::fs::File;
use std::io::BufReader;
use std::io::Read;
use std::string::String;


#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum Cell {
    Empty,
    Value(u32),
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum Direction {
    Up,
    Down,
    Left,
    Right,
    Init
}

type Board = Vec<Vec<Cell>>;

#[derive(Debug, Clone)]
struct Move {
    swap: (Cell, Cell),
    dir: Direction,
    score: u32,
    heuristic: u32,
    result: Board,
    previous_move: Option<Box<Move>>,
}

impl Ord for Move {
    fn cmp(&self, other: &Move) -> Ordering {
        //(self.score + self.heuristic).cmp(&(other.score + other.heuristic))
        (other.score + other.heuristic).cmp(&(self.score + self.heuristic))
        // (other.heuristic).cmp(&(self.heuristic))
    }
}
impl PartialOrd for Move {
    fn partial_cmp(&self, other: &Move) -> Option<Ordering> {
        Some(self.cmp(other))
    }
}
impl PartialEq for Move {
    fn eq(&self, other: &Move) -> bool {
        (self.score + self.heuristic) == (other.score + other.heuristic)
    }
}
impl Eq for Move {}

fn draw(board: &Board) {
    for line in board.iter() {
        print!("|");
        for (_i, b) in line.iter().enumerate() {
            if let Cell::Value(v) = *b {
                print!("{:5}", v);
            } else {
                print!("empty");
            }
            print!("|");
        }
        println!("");
    }
}

fn solvable(board : &Board) -> bool {

    let mut inline: Vec<Cell> = vec![];
    let mut empty = 0u32;

    for line in board.iter() {
        inline.extend(line.iter().cloned())
    }

    let mut inversion: u32 = 0;
    for i in 0..inline.len() {
        if inline[i] == Cell::Empty {
            empty = (i + 1) as u32;
        }

        for j in i..(inline.len() - 1) {
            let x = if let Cell::Value(v) = inline[i] {
                v
            } else {
                (inline.len() + 1) as u32
            };
            let y = if let Cell::Value(v) = inline[j] {
                v
            } else {
                (inline.len() + 1) as u32
            };

            // println!("{} : {}, {} : {}", i, x, j, y);
            if x > y {
                inversion += 1;
            }
        }
    }
    println!("{} {}", inversion, empty);
    return (inversion % 2) == (empty % 2);
}

fn heuristic(board : &Board) -> u32 {
    let len: u32 = (board.len() * board[0].len()) as u32;
    let mut h = 0u32;
    for (i, line) in board.iter().enumerate() {
        for (j, elem) in line.iter().enumerate() {
             let x = if let &Cell::Value(v) = elem {
                v
            } else {
                len
            };

            let ry = (x - 1) % (board.len() as u32);
            let rx = (x - 1) / (board[0].len() as u32);
            // println!("{} {} {} {} {}", x, rx, ry, i, j);
            h += (((rx as i32) - (i as i32)).abs() + ((ry as i32) - (j as i32)).abs()) as u32;
        }
    }
    return h;
}

fn next_moves(board: &Board, current_score: u32, previous_move: &Move) -> Vec<Move> {
    let mut moves: Vec<Move> = vec![];
    const DIRECTIONS: [((i8, i8), Direction); 4] = [
        ((-1, 0), Direction::Up), 
        ((0, -1), Direction::Left), 
        ((1, 0), Direction::Down), 
        ((0, 1), Direction::Right)
    ];

    let width: i8 = board.len() as i8;
    let height: i8 = board[0].len() as i8;

    let mut x: i8 = -1;
    let mut y: i8 = -1;
    for (i, line) in board.iter().enumerate() {
        for (j, b) in line.iter().enumerate() {
            if *b == Cell::Empty {
                x = i as i8;
                y = j as i8;
                break;
            }
        }
        if x >= 0 && y >= 0 {
            break;
        }
    }

    for &(dir, dir_type) in DIRECTIONS.iter() {
        let (i, j) = dir;
        let xi: i8 = x + i;
        let yj: i8 = y + j;
        if xi >= 0 && xi < width && yj >= 0 && yj < height {
            let swap: (Cell, Cell) = (board[x as usize][y as usize], board[xi as usize][yj as usize]);

            // prevent loop
            if (swap.0 == previous_move.swap.0 && swap.1 == previous_move.swap.1) || (swap.1 == previous_move.swap.0 && swap.0 == previous_move.swap.1) {
                continue;
            }


            let mut new_board = board.clone();
            new_board[x as usize][y as usize] = swap.1;
            new_board[xi as usize][yj as usize] = swap.0;
            moves.push(Move {
                swap: swap,
                dir: dir_type,
                score: current_score + 1,
                heuristic: heuristic(&new_board),
                result: new_board,
                previous_move: Some(Box::new(previous_move.clone())),
            });
        }
    }

    return moves;
}

fn hash(board: &Board) -> String {
    let mut h: String = String::new();

    for row in board.iter() {
        for cell in row.iter() {
            if let Cell::Value(x) = *cell {
                h.push_str(&x.to_string());
            } else {
                h.push('v');
            }
        }
    }

    return h;
}

fn load_board(filename: &'static str) -> Board {
    let mut board: Board = vec![];

    if let Ok(file) = File::open(filename) {
        let mut buf_reader = BufReader::new(file);
        let mut contents = String::new();

        if let Ok(_) = buf_reader.read_to_string(&mut contents) {
            for line in contents.split("\n") {
                if line.len() > 1 {
                    let mut row: Vec<Cell> = vec![];
                    for elem in line.split(" ") {
                        let x: u32 = elem.parse().unwrap();
                        if x == 0 {
                            row.push(Cell::Empty);
                        } else {
                            row.push(Cell::Value(x));
                        }
                    }
                    board.push(row);
                }
            }
        }
    }

    return board;
}

fn main() {
    let mut board: Board = vec![];
    board.push(vec![Cell::Value(1),Cell::Value(2),Cell::Value(3)]);
    board.push(vec![Cell::Value(4),Cell::Value(6),Cell::Value(8)]);
    board.push(vec![Cell::Value(7),Cell::Value(5),Cell::Empty]);
    board = load_board("data/example.txt");
    draw(&board);
    if !solvable(&board) {
        println!("This taquin is solvable !");
    } else {
        println!("This taquin is unsolvable :(");
        return;
    }

    let first_move = Move {
        swap: (Cell::Empty, Cell::Empty),
        dir: Direction::Init,
        score: 0,
        heuristic: heuristic(&board),
        result: board.clone(),
        previous_move: None,
    };

    let mut heap: BinaryHeap<Move> = BinaryHeap::new();
    let mut seen: HashSet<String> = HashSet::new();

    heap.push(first_move);
    loop {
        if let Some(m) = heap.pop() {
            if m.heuristic == 0 {
                let mut swaps: Vec<Cell> = vec![];
                let mut directions: Vec<Direction> = vec![];
                let mut last_move: Move = m.clone();
                loop {
                    swaps.push(last_move.swap.1);
                    directions.push(last_move.dir);
                    if let Some(lm) = last_move.previous_move {
                        last_move = *lm;
                    } else {
                        break;
                    }
                }
                swaps.pop();
                swaps.reverse();
                directions.pop();
                directions.reverse();

                println!("{}", seen.len());
                println!("{:?} {}", swaps, swaps.len());
                println!("{:?} {}", directions, directions.len());
                draw(&m.result);
                break;
            }
            let s: String = hash(&m.result);
            if seen.contains(&s) {
                continue;
            }
            seen.insert(s);
            //print!(".");
            // println!("{} {} {}", m.score, m.heuristic, seen.len());
            // draw(&m.result);
            //break;

            for nm in next_moves(&m.result, m.score, &m).iter() {
                let s: String = hash(&nm.result);
                if !seen.contains(&s) {
                    heap.push(nm.clone());
                }
            }
        } else {
            println!("Heap empty");
            break;
        }
    }


}
